#include <string>

#include "common.hh"
#include "debug.hh"
#include "Token.hh"
#include "Symbol.hh"
#include "RENode.hh"
#include "SRNode.hh"
#include "Rule.hh"

//
// RENode
//

RENode::RENode(SRUnion * nxt, RENode * p=0, Symbol * rw=0, SRNode * matched=0, Token * exact = 0) 
  : next(nxt), parent(p), matching(rw), matchedNode(matched), exactToken(exact)
{
  length = p
    ? (rw ? p->length+1 : p->length)
    : (rw ? 1 : 0 ) ;
  children = 0;
  copies   = 1;
}

void   RENode::addChild()          { ++children; }
void   RENode::removeChild()       { --children; }
void   RENode::addCopy()           { ++copies; }
void   RENode::removeCopy()        { --copies; }

bool   RENode::isRoot() const      { return parent   == 0; }
bool   RENode::isLeaf() const      { return children == 0; }
bool   RENode::isDead() const      { return ( children == 0 && copies == 0 ); }

RENode  * RENode::getParent() const           { return parent; }
Symbol  * RENode::getMatchingSymbol() const   { return matching; }
SRUnion * RENode::getNext() const             { return next; }

string  RENode::doString() const 
{
  return "RENode[ ";
  /*
  string thisnode =
    ((exactToken==0) 
     ? " $$$ " 
     : exactToken->toString() )
    +
    ((matching==0)
     ? " ___ "
     : ((matching==Symbol::SKIP_SYMBOL) 
	? " <=> ..."
	: " <=> " + matching->toString())
     );
  
  return
    ((parent)
     ? parent->doString() + "; " + thisnode
     : "RENode[ " + thisnode );
  */
  /*
  string thisnode = 
    ((parent && getNextToken()==0) 
     ? " _$_" 
     : getNextToken()->toString() ) ;
  
  return
    (parent 
     ? parent->doString() +
     (parent->matching 
      ? " => " + parent->matching->toString() + " ; " + thisnode 
      : "; ")
     :  "RENode[ " + getNextToken()->toString()) ;
  */
}

string  RENode::toString() const 
{
  string s =
    doString() + " ] (" + 
    (isDead()?"dead":"live") + " " + 
    (isLeaf()?"leaf":"inner") + ")";
  
  return s;
}

TokenVector * RENode::doAccept(Rule * rule, 
			     string & X, string & x, 
			     string & Y, string & y) const
{
  TokenVector * tokens = 
    parent 
    ? parent->doAccept(rule,X,x,Y,y)
    : new TokenVector() ;
  
  if(matching)
    {
      X = matching->funcs ->first()=="X" ? exactToken->func  : X ;
      x = matching->forms ->first()=="x" ? exactToken->form  : x ;
      Y = matching->sfuncs->first()=="Y" ? exactToken->sfunc : Y ;
      y = matching->sforms->first()=="y" ? exactToken->sform : y ;
      
      Token * t = matching->rewrite(exactToken);
      tokens->insert(tokens->end(), t);
    }
  
  return tokens;
}

/**
 * Rewrite the matched string to a single SRNode.
 */
SRNode * RENode::reduce(Rule * rule) const
{
  string 
    X = rule->left->funcs ->first(),
    x = rule->left->forms ->first(),
    Y = rule->left->sfuncs->first(),
    y = rule->left->sforms->first();
  
  TokenVector * tv = doAccept(rule, X,x,Y,y);
  
  string fu,fo,sfu,sfo;
  
  fu = rule->left->funcs->first()=="X" ? X : rule->left->funcs ->first() ;
  if(rule->left->funcs->first()=="Y") fu = Y;
  fo = rule->left->forms->first()=="x" ? x : rule->left->forms->first() ;
  if(rule->left->forms->first()=="y") fo = y;
  sfu = rule->left->sfuncs->first()=="Y" ? Y : rule->left->sfuncs ->first() ;
  sfo = rule->left->sforms->first()=="y" ? y : rule->left->sforms->first() ;
  
  Token * token = new TokenGroup(tv, fu,fo,sfu,sfo);
  return new SRNode(token, next);
}
