#ifndef _Node_hh_
#define _Node_hh_

#include <string>

#include "common.hh"

class RENode;
class SRNode;
class SRUnion;
class Token;
class TokenGroup;
class Symbol;

/**
 * RENode: A node in the Regular Expression parser tree.
 */
class RENode
{
private:
  /** Children */
  unsigned int children;
  /** Live Copies */
  unsigned int copies;
  
  /** Defines next token in inputstring. */
  SRUnion * next;
  /** Parent node. */
  RENode * parent;
  /** Matching symbol. */
  Symbol * matching;
  /** The SRNode that contains the matched token. */
  SRNode * matchedNode;
  /** The exact token matched (i.e. never a TokenSet) */
  Token * exactToken;
  
  /** # of matched tokens in this state. */
  unsigned int length;
  
  TokenVector * doAccept(Rule * rule, 
			 string & X, string & x, 
			 string & Y, string & y) const;
  
  string doString() const;
  
public:
  bool isLeaf() const;
  bool isRoot() const;
  bool isDead() const;
  
  void addChild();
  void removeChild();
  void addCopy();
  void removeCopy();
  
  RENode  * getParent() const;
  Symbol  * getMatchingSymbol() const;
  SRUnion * getNext() const;
  SRNode  * getMatchedNode() const { return matchedNode; }
  
  SRNode * reduce(Rule * rule) const;
  
  string toString() const;
  
  RENode(SRUnion * nxt, RENode * p=0, Symbol * rw=0, SRNode * matched=0, Token * exact=0);
};

#endif
