#ifndef _Regex_hh_
#define _Regex_hh_

#include <string>

class Tree;
class Symbol;

class RegExpr
{
public:
  static unsigned int matchInvoked;
  RegExpr(){}
  
  virtual void match(Tree*) const;
  virtual string toString() const;
};

class LiteralExpr: public RegExpr
{
private:
  Symbol * literal;
  
public:
  LiteralExpr(Symbol* sym) : 
    literal(sym)
    {}
  
  virtual void match(Tree*) const;
  virtual string toString() const;
};

class SequenceExpr: public RegExpr
{
private:
  RegExpr * expr1;
  RegExpr * expr2;
  
public:
  SequenceExpr(RegExpr* re1, RegExpr* re2) :
    expr1(re1), expr2(re2)
    {}
  
  virtual void match(Tree*) const;
  virtual string toString() const;
};

class RepetitionExpr: public RegExpr
{
private:
  RegExpr * expr;
  
public:
  RepetitionExpr(RegExpr * re) :
    expr(re)
    {}
  
  virtual void match(Tree*) const;
  virtual string toString() const;
};

class OptionalExpr: public RegExpr
{
private:
  RegExpr * expr;
  
public:
  OptionalExpr(RegExpr * re) :
    expr(re)
    {}
  
  virtual void match(Tree*) const;
  virtual string toString() const;
};

class AlternationExpr: public RegExpr
{
private:
  RegExpr * expr1;
  RegExpr * expr2;
  
public:
  AlternationExpr(RegExpr* re1, RegExpr* re2) : 
    expr1(re1), expr2(re2)
    {}
  
  virtual void match(Tree*) const;
  virtual string toString() const;
};

#endif
