#include "common.hh"
#include "debug.hh"
#include "compiler.hh"
#include "Token.hh"
#include "SRNode.hh"
#include "Rule.hh"
#include <vector>
#include <string>
#include <map>


//engcg | /home/eckhard/perl/engfilter.pre | mdis -x English --grammar ~eckhard/english/engrules | /home/eckhard/perl/addclausefunction | /home/eckhard/perl/niceline.perl | /home/eckhard/perl/engfilter.psg@# | ./bin/cg2tree ./src/psg-grammar3

extern Symbol * const SKIP_SYMBOL;
extern Rules * rules;
extern int tracking;
extern int tree;
extern int printComplete;
extern int printIncompletable;
extern int printAllStates;
extern int experimental;

vector<Token*> * getGreedyGuess(SRUnion * uni, vector<Token*> * group = 0)
{
  if(group==0) group = new vector<Token*>();
  if(uni->getNodes()==0) return group;
  SRNode * greediest = 0;
  unsigned int maxGreed = 0;
  for(list<SRNode*>::iterator 
	i = uni->getNodes()->begin(),
	z = uni->getNodes()->end();
      i!=z; ++i)
    {
      SRNode * node = *i;
      unsigned int greed = node->getToken()->getSize();
      if(greed > maxGreed) {
	maxGreed = greed;
	greediest = node;
      }
    }
  group->push_back(greediest->getToken());
  group = getGreedyGuess(greediest->getNext(), group);
  return group;
}

vector<Token*> * getTree(Token * token, vector<Token*> * group = 0)
{
  if(token==0) {
    cerr << "token==0 in getTree()\n";
    return group;
  }
  if(group==0) group = new vector<Token*>();
  Token * prev = token->getPrevious();
  if(prev != 0) getTree(prev, group);
  group->push_back(token);
  return group;
}

void printForest(SRUnion * uni, unsigned int & nPartial)
{
  if(uni->getNodes()==0) return;
  //cout << uni << "\n";
  for(list<SRNode*>::iterator 
	i = uni->getNodes()->begin(),
	z = uni->getNodes()->end();
      i!=z; ++i)
    {
      SRNode * node = *i;
      //cout << node << " " << node->toString()<<"\n";
      //cout << node << " " << node->getNext()->getNodes() << "\n";
      if(node->getNext()->getNodes()==0) {
	if(nPartial > 0 ) cout <<"&&\n";
	++nPartial;
	cout << "A" << nPartial << "\n";
	vector<Token*> * group = getTree(node->getToken());
	Token * partial = new TokenGroup(group, "PARTIAL TREE", "The rules could not construct a complete tree");
	cout << partial->toVISL(0,true);
	delete partial;
	delete group;
      }
      else {
	//cout << node->getNext() << "*\n";
	printForest(node->getNext(), nPartial);
      }
    }
}

vector<Token*> * getForest(SRUnion * uni, vector<Token*> * group=0, vector<Token*> * forest=0)
{
  if(uni==0)
    {
      cout << "uni==0!\n";
      return forest;
    }
  if(group==0) group = new vector<Token*>();
  if(forest==0) forest = new vector<Token*>();
  
  if(uni->getNodes()==0)
    {
      forest->push_back(new TokenGroup(group, "PARTIAL TREE", "The rules could not construct a complete tree"));
      return forest;
    }
  
  for(list<SRNode*>::iterator 
	i = uni->getNodes()->begin(),
	z = uni->getNodes()->end();
      i!=z; ++i)
    {
      vector<Token*> * pt = new vector<Token*>(*group);
      pt->push_back((*i)->getToken());
      getForest((*i)->getNext(), pt, forest);
    }
  return forest;
}

/**
 * Best and most recent algorithm (with new data structures + shift unification optimization). 
 * Builds right-to-left. Initial stateset has one state: 
 * the startstate with the cursor to the far right.
 */
void doParse(Token * endToken)
{
  debug(1) << "\n----------------\n";
  
  vector<SRUnion*> * unions = new vector<SRUnion*>();
  SRUnion * unified = new SRUnion();
  unions->push_back(unified);
  Token * token = endToken;
  while(token)
    {
      debug(1) << "Token: " << token->toString() << "\n";
      
      list<SRNode*>* outputStates = new list<SRNode*>();
      list<SRNode*>* activeStates = new list<SRNode*>();
      
      // create initial state at this position.
      SRNode * shifted = new SRNode(token, unified);
      activeStates->push_back(shifted);
      outputStates->push_back(shifted);
      
      //while there are any active states:
      while(activeStates->size() > 0)
	{
	  debug(1) << activeStates->size() << " active states...\n";
	  
	  //  this list will be the list of active states in the next iteration
	  list<SRNode*> * newActiveStates = new list<SRNode*>();
	  
	  // to test for uniqueness of matches
	  map<string,bool>* unique = new map<string,bool>();
	  
	  //  for each rule:
	  for(Rules::iterator ri=rules->begin(), rz=rules->end(); 
	      ri!=rz; ++ri)
	    {
	      Rule * rule = (*ri);
	      
	      //      find matches for this rule.
	      list<SRNode*> * matchingStates = rule->match(activeStates);
	      if(matchingStates->size()>0) debug(1) << rule->toString() << " produced " << matchingStates->size() << " states:\n";
	      
	      //      iterate through each matching state:
	      for(list<SRNode*>::iterator
		    i=matchingStates->begin(),
		    z=matchingStates->end();
		  i!=z; ++i)
		{
		  SRNode * matching = *i;
		  string s = matching->getToken()->toString();
		  matching->getToken()->setPrevious(token->getPrevious());
		  
		  // test '+' rule modifier (never use this rule to create a token with < 2 children)
		  if(rule->modifier=='+' && matching->getToken()->getSize() < 2) 
		    {
		      debug(1) << "  " << s << "[HAS TOO FEW CHILDREN]\n";
		      delete matching;
		      continue;
		    }
		  
		  // test '!' rule modifier (never use this rule to create a root node)
		  if(rule->modifier=='!' && matching->isComplete())
		    {
		      debug(1) << "  " << s << "[IS ROOT]\n";
		      delete matching;
		      continue;
		    }
                  
		  // test uniqueness of match. 
		  if(!((*unique)[s]))
		    (*unique)[s] = true;
		  else
		    {
		      debug(1) << "  " << s << "[IS DUPLICATE]\n";
		      delete matching;
		      continue;
		    }
		  
		  // test against impossible discontinuities.
		  if(matching->getToken()->hasImpossibleDiscontinuities())
		    {
		      debug(1) << "  " << s << "[HAS INCOMPLETE DISCONTINUITY]\n";
		      delete matching;
		      continue;
		    }
		  
		  // accept the matching state.
		  newActiveStates->push_back(matching);
		  outputStates->push_back(matching);
		  debug(1) << "  " << s << "\n";
		}
	      
	      delete matchingStates;
	    }
	  delete unique;
	  delete activeStates;
	  activeStates = newActiveStates;
	}
      debug(1) << "there are " << outputStates->size() << " states at this point.\n";
      
      unified = new SRUnion(outputStates);
      unions->push_back(unified);
      token = token->getPrevious();
      debug(1) << "----------------\n";
    }
  
  if(tree && unified->getNodes())
    {
      // Print all successfully built trees
      unsigned int nCompleted = 0;
      for(list<SRNode*>::iterator 
	    i=unified->getNodes()->begin(),
	    z=unified->getNodes()->end();
	  i!=z; ++i)
	{
	  if((*i)->isComplete())
	    {
	      if(nCompleted > 0 ) cout <<"&&\n";
	      ++nCompleted;
	      cout << "A" << nCompleted << "\n";
	      cout << (*i)->getToken()->toVISL(0,true);
	    }
	}
      
      // If there were no successes, use unparsed tokens.
      if(nCompleted==0)
	{
	  Token * guess1 = new TokenGroup(getGreedyGuess(unified), "PARTIAL TREE", "The rules could not construct a complete tree");
	  cout << "A1\n";
	  cout << guess1->toVISL(0, true);
	  
	  map<SRUnion*, unsigned int> * umap = new map<SRUnion*, unsigned int>();
	  map<SRNode*, unsigned int> * nmap = new map<SRNode*, unsigned int>();
	  map<SRNode*, SRUnion*> * numap = new map<SRNode*, SRUnion*>();
	  map<SRUnion*, vector<SRNode*>*> * incoming = new map<SRUnion*, vector<SRNode*>*>();
	  unsigned int nUnions = 0;
	  
	  for(vector<SRUnion*>::iterator
		ui=unions->begin(),
		uz=unions->end();
	      ui!=uz; ++ui)
	    {
	      SRUnion * uni = *ui;
	      (*umap)[uni] = nUnions;
	      (*incoming)[uni] = new vector<SRNode*>();
	      ++nUnions;
      
	      if(uni->getNodes()==0) continue;
	      for(list<SRNode*>::iterator 
		    ni=uni->getNodes()->begin(),
		    nz=uni->getNodes()->end();
		  ni!=nz; ++ni)
		{
		  SRNode * node = *ni;
		  SRUnion * next = node->getNext();
		  (*numap)[node] = uni;
		  (*nmap)[node] = nUnions - (*umap)[next];
		  ((*incoming)[next])->push_back(node);
		}
	    }
	  
	  SRUnion * theUnion = *(unions->begin());
	  vector<Token*> * theTree = new vector<Token*>();
	  
	  while(((*incoming)[theUnion])->size() > 0)
	    {
	      unsigned int best = 0;
	      SRNode * bestNode = 0;
	      for(vector<SRNode*>::iterator
		    i=((*incoming)[theUnion])->begin(),
		    z=((*incoming)[theUnion])->end();
		  i!=z; ++i)
		{
		  SRNode * node = *i;
		  if(best < (*nmap)[node])
		    {
		      best = (*nmap)[node];
		      bestNode = node;
		    }
		}
	      theUnion = (*numap)[bestNode];
	      theTree->push_back(bestNode->getToken());
	    }
	  
	  vector<Token*> * theTree2 = new vector<Token*>();
	  theTree2->insert(theTree2->begin(), theTree->rbegin(), theTree->rend());
	  Token * guess2 = new TokenGroup(theTree2, "PARTIAL TREE", "The rules could not construct a complete tree");  
	  cout << "&&\nA2\n";
	  cout << guess2->toVISL(0, true);
	}
    }
}
